import tensorflow as tf
from tensorflow.keras.initializers import Initializer, HeUniform, HeNormal
from tensorflow.keras.layers import Layer, Dropout, LayerNormalization, BatchNormalization, UnitNormalization


class GridInitializer(Initializer):
    """Initializes a grid for use in B-spline calculations within KANLinear layer.

    Args:
        grid_range (list of float): The min and max range values for the grid.
        grid_size (int): The number of intervals in the grid.
        spline_order (int): The order of the spline transformation.
    """
    def __init__(self, grid_range, grid_size, spline_order):
        self.grid_range = grid_range
        self.grid_size = grid_size
        self.spline_order = spline_order

    def __call__(self, shape, dtype=None):
        h = (self.grid_range[1] - self.grid_range[0]) / self.grid_size
        grid = tf.linspace(
            start=-self.spline_order * h + self.grid_range[0],
            stop=(self.grid_size + self.spline_order) * h + self.grid_range[0],
            num=self.grid_size + 2 * self.spline_order + 1
        )
        grid = tf.tile(tf.expand_dims(tf.expand_dims(grid, 0), 0), [1, shape[1], 1])
        return grid

    def get_config(self):
        return {
            "grid_range": self.grid_range,
            "grid_size": self.grid_size,
            "spline_order": self.spline_order
        }
        
class KANLinear(Layer):
    """Custom Keras layer that implements Kernel Additive Networks (KAN) with B-spline transformations.

    This layer supports tensors with more than two dimensions, treating the last dimension as features
    and applies transformations uniformly across any preceding dimensions (e.g., time steps in sequences).

    Do not need in_features, computed at build time

    Args:
        units (int): Number of output units.
        grid_size (int, optional): Size of the grid for spline calculations. Defaults to 5.
        spline_order (int, optional): Order of the spline transformations. Defaults to 3.
        scale_noise (float, optional): Scaling factor for noise. Defaults to 0.1.
        scale_base (float, optional): Base scaling factor. Defaults to 1.0.
        base_activation (str, optional): Activation function to use. Defaults to 'silu'.
        grid_range (list of float, optional): Range of values for the grid. Defaults to [-1, 1].
    """
    def __init__(
        self,
        units,
        grid_size=3,
        spline_order=3,
        base_activation='silu',
        grid_range=[-1, 1],
        dropout = 0.,
        use_bias = True,
        use_layernorm = True,
        **kwargs
    ):
        super(KANLinear, self).__init__(**kwargs)
        self.in_features = None
        self.units = units
        self.grid_size = grid_size
        self.spline_order = spline_order

        self.base_activation = getattr(tf.nn, base_activation)
        self.grid_range = grid_range
        self.use_bias = use_bias
        self.use_layernorm = use_layernorm
        if self.use_layernorm:
            self.layer_norm = LayerNormalization()

        self.dropout = Dropout(dropout)
    
    def build(self, input_shape):
        super(KANLinear, self).build(input_shape)
        self.in_features = input_shape[-1]
        self.other_dims = input_shape[1:-1]

        self.grid = self.add_weight(
            name="grid",
            shape=[1, self.in_features, self.grid_size + 2 * self.spline_order + 1],
            initializer=GridInitializer(self.grid_range, self.grid_size, self.spline_order),
            trainable=False
        )

        self.base_weight = self.add_weight(
            name="base_weight",
            shape=[self.units, self.in_features],
            initializer=HeUniform
        )

        if self.use_bias:
            self.base_bias = self.add_weight(
                name="base_bias",
                shape=[self.units,],
                initializer="zeros"
            )

        self.spline_weight = self.add_weight(
            name="spline_weight",
            shape=[self.units, self.in_features * (self.grid_size + self.spline_order)],
            initializer=HeUniform
        )

    def call(self, x):
        if self.use_layernorm:
            x = self.layer_norm(x)
        if self.use_bias:
            base_output = tf.matmul(self.base_activation(x), self.base_weight, transpose_b=True) + self.base_bias
        else:
            base_output = tf.matmul(self.base_activation(x), self.base_weight, transpose_b=True)
        spline_output = tf.matmul(
            self.b_splines(x),
            self.spline_weight,
            transpose_b=True
        )
        return self.dropout(base_output) + self.dropout(spline_output)
            
    def b_splines(self, x):
        batch_size = tf.shape(x)[0]
        x_expanded = tf.expand_dims(x, -1)  
        
        grid_expanded = tf.broadcast_to(self.grid, (batch_size, self.in_features, self.grid.shape[2]))
        done_dims = []
        for dim in self.other_dims:
            grid_expanded = tf.expand_dims(grid_expanded, 1) 
            grid_expanded = tf.broadcast_to(grid_expanded, (batch_size, dim, *done_dims, self.in_features, self.grid.shape[2]))
            done_dims.append(dim)

        bases = tf.cast((x_expanded >= grid_expanded[..., :-1]) & (x_expanded < grid_expanded[..., 1:]), x.dtype)
        for k in range(1, self.spline_order + 1):
            left_denominator = grid_expanded[..., k:-1] - grid_expanded[..., :-(k + 1)]
            right_denominator = grid_expanded[..., k + 1:] - grid_expanded[..., 1:-k]
            
            left = (x_expanded - grid_expanded[..., :-(k + 1)]) / left_denominator
            right = (grid_expanded[..., k + 1:] - x_expanded) / right_denominator
            bases = left * bases[..., :-1] + right * bases[..., 1:]
            
        bases = tf.reshape(bases, (batch_size, *bases.shape[1:-2], -1))
        return bases

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.units,)




import tensorflow as tf
from tensorflow.keras.layers import Layer, Dense, Add, LayerNormalization, Multiply, Dropout

from kamoe import KANLinear

class AddAndNorm(Layer):
    def __init__(self, **kwargs):
        super(AddAndNorm, self).__init__(**kwargs)
        self.add_layer = Add()
        self.norm_layer = LayerNormalization()
    
    def call(self, inputs):
        tmp = self.add_layer(inputs)
        tmp = self.norm_layer(tmp)
        return tmp

    def compute_output_shape(self, input_shape):
        return input_shape[0]  # Assuming all input shapes are the same

class Gate(Layer):
    def __init__(self, hidden_layer_size = None, **kwargs):
        super(Gate, self).__init__(**kwargs)
        self.hidden_layer_size = hidden_layer_size

    def build(self, input_shape):
        if self.hidden_layer_size is None:
            self.hidden_layer_size = input_shape[-1]
        self.dense_layer = Dense(self.hidden_layer_size)
        self.gated_layer = Dense(self.hidden_layer_size, activation='sigmoid')
        super(Gate, self).build(input_shape)

    def call(self, inputs):
        dense_output = self.dense_layer(inputs)
        gated_output = self.gated_layer(inputs)
        return Multiply()([dense_output, gated_output])

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.hidden_layer_size,)


class GRKAN(Layer):
    def __init__(self, hidden_layer_size, output_size=None, activation = None, dropout = 0.1, use_bias = False, **kwargs):
        super(GRKAN, self).__init__(**kwargs)
        self.hidden_layer_size = hidden_layer_size
        self.output_size = output_size
        self.activation = tf.keras.activations.get(activation) if activation is not None else None
        self.dropout = Dropout(dropout)
        self.dropout_value = dropout
        self.use_bias = use_bias

    def build(self, input_shape):
        if self.output_size is None:
            self.output_size = self.hidden_layer_size
        self.skip_layer = Dense(self.output_size)
        
        self.hidden_layer_1 = KANLinear(self.hidden_layer_size, base_activation='elu', dropout = self.dropout_value, use_bias = self.use_bias, use_layernorm=False)
        self.hidden_layer_2 = KANLinear(self.hidden_layer_size, dropout = self.dropout_value, use_bias = self.use_bias, use_layernorm=False)
        self.gate_layer = Gate(self.output_size)
        self.add_and_norm_layer = AddAndNorm()
        super(GRKAN, self).build(input_shape)

    def call(self, inputs):
        if self.skip_layer is None:
            skip = inputs
        else:
            skip = self.skip_layer(inputs)
        
        hidden = self.hidden_layer_1(inputs)
        hidden = self.hidden_layer_2(hidden)
        hidden = self.dropout(hidden)
        gating_output = self.gate_layer(hidden)
        output = self.add_and_norm_layer([skip, gating_output])
        if self.activation is not None:
            output = self.activation(output)
        return output

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.output_size,)

class GRN(Layer):
    def __init__(self, hidden_layer_size, output_size=None, activation = None, dropout = 0.1, **kwargs):
        super(GRN, self).__init__(**kwargs)
        self.hidden_layer_size = hidden_layer_size
        self.output_size = output_size
        self.activation = tf.keras.activations.get(activation) if activation is not None else None
        self.dropout = Dropout(dropout)
        self.dropout_value = dropout

    def build(self, input_shape):
        if self.output_size is None:
            self.output_size = self.hidden_layer_size
        self.skip_layer = Dense(self.output_size)
        
        self.hidden_layer_1 = Dense(self.hidden_layer_size, activation='elu')
        self.hidden_layer_2 = Dense(self.hidden_layer_size)
        self.gate_layer = Gate(self.output_size)
        self.add_and_norm_layer = AddAndNorm()
        super(GRN, self).build(input_shape)

    def call(self, inputs):
        if self.skip_layer is None:
            skip = inputs
        else:
            skip = self.skip_layer(inputs)
        
        hidden = self.hidden_layer_1(inputs)
        hidden = self.hidden_layer_2(hidden)
        hidden = self.dropout(hidden)
        gating_output = self.gate_layer(hidden)
        output = self.add_and_norm_layer([skip, gating_output])
        if self.activation is not None:
            output = self.activation(output)
        return output

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.output_size,)


import tensorflow as tf
from tensorflow.keras.layers import Layer, Dense, Dropout, RNN
from tensorflow.keras.models import Sequential, clone_model
from kamoe import GRKAN, GRN
import logging

logger = logging.getLogger(__name__)

class MoE(Layer):
    def __init__(self, base_expert, n_experts=4, gating_network_activation = 'sigmoid', dropout=0., **kwargs):
        super().__init__(**kwargs)
        self.n_experts = n_experts
        self.gating_network_activation = gating_network_activation
        self.experts = [clone_model(base_expert) for _ in range(n_experts)]
        self.base_expert = self.experts[0]
        self.dropout = Dropout(dropout)
        self.input_is_sequence = None
        self.output_is_sequence = None
        self.flatten_input = None
        self.output_units = None
        self.gating_network_cls = GRN

    def build(self, input_shape):
        self.input_is_sequence = len(input_shape) == 3

        # Build one expert to determine output shape
        test_input = tf.keras.Input(shape=input_shape[1:])
        test_output = self.base_expert(test_input)
        self.output_shape = test_output.shape
        self.output_is_sequence = len(self.output_shape) == 3
        self.output_units = self.output_shape[-1]

        self.hidden_layer = Dense(self.output_units, 'relu')
        

        if self.input_is_sequence:
            if self.output_is_sequence:
                self.flatten_input = False
                weight_input_shape = input_shape
            else:
                # Check if there's an RNN layer with return_sequences=False
                if isinstance(self.base_expert, Sequential):
                    rnn_layers = [layer for layer in self.base_expert.layers if hasattr(layer, 'return_sequences')] # Using attribute return sequence over RNN type to covers wider cases
                    last_rnn_layer = next((layer for layer in reversed(rnn_layers) if not layer.return_sequences), None)
                    self.flatten_input = last_rnn_layer is None
                elif isinstance(self.base_expert, Layer):
                    if hasattr(self.base_expert, 'return_sequences'):
                        if self.base_expert.return_sequences:
                            raise ValueError('Base expert has return_sequences=True but output is not a sequence')
                        self.flatten_input = False
                    else:
                        logger.warning('Base expert does not have return_sequences attribute. Assuming it uses flattening inside')
                        self.flatten_input = True
                else:
                    logger.warning('Base expert does not have return_sequences attribute. Assuming it uses flattening inside')
                    self.flatten_input = True

                if self.flatten_input:
                    weight_input_shape = (*input_shape[:-2], input_shape[-2] * input_shape[-1])
                else:
                    weight_input_shape = (*input_shape[:-2], input_shape[-1])
        else:
            self.flatten_input = False
            weight_input_shape = input_shape

        self.hidden_layer.build(weight_input_shape)
        self.hidden_to_weight = self.gating_network_cls(self.n_experts, activation=self.gating_network_activation, dropout=self.dropout.rate)
        self.hidden_to_weight.build(self.hidden_layer.compute_output_shape(weight_input_shape))

        for expert in self.experts:
            if not expert.built: 
                expert.build(input_shape)
        super().build(input_shape)

    def call(self, inputs):
        if self.input_is_sequence:
            if self.flatten_input:
                hidden = self.hidden_layer(tf.reshape(inputs, (tf.shape(inputs)[0], -1)))
            elif not self.output_is_sequence:
                hidden = self.hidden_layer(inputs[:, -1, :])
            else:
                hidden = self.hidden_layer(inputs)
        else:
            hidden = self.hidden_layer(inputs)

        weights = self.hidden_to_weight(hidden)
        expert_outputs = tf.stack([expert(inputs) for expert in self.experts], axis=-1)

        if self.output_is_sequence:
            weighted_outputs = expert_outputs * weights[..., tf.newaxis, :]
        else:
            weighted_outputs = expert_outputs * weights[..., tf.newaxis, :]

        return tf.reduce_sum(weighted_outputs, axis=-1)

class KAMoE(MoE):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gating_network_cls = GRKAN

